import requests
import requests.adapters

import uuid

from urllib3.util.retry import Retry
from typing import List

from vulnebify.errors import *
from vulnebify.models import *


class Idempotency:
    @staticmethod
    def key() -> str:
        return f"ik_{uuid.uuid4().hex}"


class VulnebifyHttpClient:
    def __init__(
        self,
        base_url: str,
        api_key: str,
        proxy_url: str | None = None,
    ):
        self.__base_url = base_url.rstrip("/")
        self.__api_key = api_key
        self.__proxy_url = proxy_url
        self.__session = None

    def get(self, path: str, **kwargs) -> str:
        try:
            self.__ensure_session_initialized()

            response = self.__session.get(f"{self.__base_url}{path}", **kwargs)

            if response.status_code != 200:
                raise VulnebifyApiError(response.status_code, response.text)

            return response.text
        except VulnebifyApiError:
            raise
        except Exception as e:
            raise VulnebifyClientError(e)

    def post(self, path: str, json=None, idempotency_key=None, **kwargs) -> str:
        try:
            self.__ensure_session_initialized()

            headers = kwargs.pop("headers", {})

            headers["Idempotency-Key"] = idempotency_key or Idempotency.key()

            response = self.__session.post(
                f"{self.__base_url}{path}", json=json, headers=headers, **kwargs
            )

            if response.status_code != 200:
                raise VulnebifyApiError(response.status_code, response.text)

            return response.text
        except VulnebifyApiError:
            raise
        except Exception as e:
            raise VulnebifyClientError(e)

    def __create_session(self):
        session = requests.Session()
        session.headers.update({"Authorization": f"Bearer {self.__api_key}"})

        if self.__proxy_url:
            session.proxies.update(
                {
                    "http": self.__proxy_url,
                    "https": self.__proxy_url,
                }
            )

        retries = Retry(
            total=5,
            backoff_factor=1.0,
            status_forcelist=[429, 500, 502, 503, 504],
            allowed_methods=["GET", "HEAD", "OPTIONS", "POST"],
            raise_on_status=False,
        )
        adapter = requests.adapters.HTTPAdapter(max_retries=retries)

        session.mount("http://", adapter)
        session.mount("https://", adapter)

        return session

    def __enter__(self):
        self.__session = self.__create_session()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__session.close()
        self.__session = None

    def __ensure_session_initialized(self):
        if self.__session:
            return

        raise RuntimeError(
            "Session not initialized. Initialize 'client = VulnebifyHttpClient(...)' and use 'with client as c:'."
        )


class VulnebifyKey:
    def __init__(self, client: VulnebifyHttpClient):
        self.__client = client

    def generate(self) -> GeneratedKey:
        with self.__client as client:
            response = client.post(f"/key")

            return GeneratedKey.model_validate_json(response)

    def is_activated(self) -> bool:
        with self.__client as client:
            try:
                _ = client.get(f"/key")
                return True
            except VulnebifyApiError as e:
                if e.status_code == 401:
                    return False
                raise


class VulnebifyDomain:
    def __init__(self, client: VulnebifyHttpClient):
        self.__client = client

    def get(self, domain: str):
        with self.__client as client:
            response = client.get(f"/domain/{domain}")

            return RootDomain.model_validate_json(response)


class VulnebifyHost:
    def __init__(self, client: VulnebifyHttpClient):
        self.__client = client

    def get(self, ip_str: str):
        with self.__client as client:
            response = client.get(f"/host/{ip_str}")

            return Host.model_validate_json(response)


class VulnebifyScan:
    def __init__(self, client: VulnebifyHttpClient):
        self.__client = client

    def list(self) -> ScanList:
        with self.__client as client:
            response = client.get(f"/scan")
            return ScanList.model_validate_json(response)

    def run(
        self,
        scopes: List[str],
        ports: List[str | int],
        scanners: List[str],
        idempotency_key: str | None = None,
    ) -> str:
        with self.__client as client:
            request = {"scopes": scopes, "ports": ports, "scanners": scanners}

            response = client.post(f"/scan/", request, idempotency_key=idempotency_key)

            return ScanRun.model_validate_json(response).scan_id

    def get(self, scan_id: str) -> Scan:
        with self.__client as client:
            response = client.get(f"/scan/{scan_id}")

            return Scan.model_validate_json(response)

    def cancel(self, scan_id: str):
        with self.__client as client:
            client.post(f"/scan/{scan_id}/cancel")


class VulnebifyScanner:
    def __init__(self, client: VulnebifyHttpClient):
        self.__client = client

    def list(self) -> ScannerList:
        with self.__client as client:
            response = client.get("/scanner")

            return ScannerList.model_validate_json(response)


class Vulnebify:
    def __init__(
        self,
        api_key: str,
        api_url: str = "https://api.vulnebify.com/v1",
        proxy_url: str | None = None,
    ):
        client = VulnebifyHttpClient(api_url, api_key, proxy_url)

        self.domain = VulnebifyDomain(client)
        self.key = VulnebifyKey(client)
        self.host = VulnebifyHost(client)
        self.scan = VulnebifyScan(client)
        self.scanner = VulnebifyScanner(client)



import json

from typing import Dict, Any


class VulnebifyError(Exception):
    def __init__(self, message: str):
        self.message = message


class VulnebifyClientError(VulnebifyError):
    def __init__(self, message: str):
        super().__init__(message)


class VulnebifyApiError(VulnebifyError):
    def __init__(self, status_code: int, response: str):
        self.status_code = status_code

        try:
            self.response: Dict[str, Any] = json.loads(response)
        except:
            self.response: Dict[str, Any] = {
                "error": {"code": "processing_error", "message": response}
            }

        self.message = f"Oh no! API '{status_code}' error."



import json
import sys

from abc import ABC, abstractmethod

from vulnebify.models import *
from vulnebify.errors import VulnebifyError, VulnebifyApiError


class OutputType(str, Enum):
    HUMAN = "human"
    JSON = "json"


class Output(ABC):
    @abstractmethod
    def print_active_api_key(self, api_key: str):
        raise NotImplementedError()

    @abstractmethod
    def print_inactive_api_key(self, api_key: str, api_key_hash: str):
        raise NotImplementedError()

    @abstractmethod
    def print_host(self, host: Host):
        raise NotImplementedError()

    @abstractmethod
    def print_domain(self, domain: RootDomain):
        raise NotImplementedError()

    @abstractmethod
    def print_scan_list(self, scans: ScanList):
        raise NotImplementedError()

    @abstractmethod
    def print_scan(self, scan: Scan):
        raise NotImplementedError()

    @abstractmethod
    def print_scan_initiating(
        self,
        scopes: List[str],
        ports: List[int | str],
        scanners: List[str],
    ):
        raise NotImplementedError()

    @abstractmethod
    def print_scan_initiated(self, scan_id: str):
        raise NotImplementedError()

    @abstractmethod
    def print_scan_progress(self, scan: Scan, last_seen: datetime) -> datetime:
        raise NotImplementedError()

    @abstractmethod
    def print_scan_cancel(self, scan_id: str):
        raise NotImplementedError()

    @abstractmethod
    def print_scanner_list(self, scanners: ScannerList):
        raise NotImplementedError()

    @abstractmethod
    def print_error(self, error: VulnebifyError):
        raise NotImplementedError()

    @abstractmethod
    def print_unexpected_error(self, error: Exception):
        raise NotImplementedError()

    @abstractmethod
    def print_exit(self):
        raise NotImplementedError()


class HumanOutput(Output):
    def print_active_api_key(self, _: str):
        message = """⚠️  Use Vulnebify responsibly. Only scan systems you own or have explicit permission to test.
Violation may lead to legal consequences. See terms and conditions: https://vulnebify.com/terms

✅ API key saved successfully!"""
        print(message)

    def print_inactive_api_key(self, api_key_hash: str):
        message = f"""
To activate your API key, visit the following URL in your browser:
🔗 https://vulnebify.com/login/?api_key_hash={api_key_hash}
"""
        print(message)

    def print_host(self, host: Host):
        print(f"IP: {host.ip_str}")

        if host.location or host.autonomous_system:
            print("")

        if host.location:
            location = [host.location.country, host.location.region, host.location.city]
            joined_location = ", ".join([l for l in location if l])
            print(f"Location: {joined_location}")

        if host.autonomous_system:
            print(
                f"Organization: {host.autonomous_system.name} (ASN: {host.autonomous_system.asn})"
            )

        if host.hostnames or host.cves:
            print("")

        if host.hostnames:
            hostnames = ", ".join(host.hostnames)
            print(f"Hostnames: {hostnames}")

        if host.cves:
            cves = ", ".join(host.cves)
            print(f"CVEs: {cves}")

        if host.services:
            print("")
            print(f"=== Ports ===")
            for port in host.services:
                print(
                    f"{port.protocol.value.upper()} {port.port}/{port.transport.value.upper()}"
                )
                lines = port.banner.splitlines()
                for line in lines:
                    print(f"    {line}")
                print("")

    def print_domain(self, domain: RootDomain):
        print(f"Domain: {domain.domain}")

        if domain.dns:
            print("")
            records = ", ".join([f"{r.value} ({r.type.value})" for r in domain.dns])
            print(f"DNS: {records}")

        if domain.subdomains:
            print("")
            for subdomain in domain.subdomains:
                records = ", ".join(
                    [f"{r.value} ({r.type.value})" for r in subdomain.dns]
                )
                if records:
                    print(f"{subdomain.domain}: {records}")
                else:
                    print(f"{subdomain.domain}")

    def print_scan_list(self, scans: ScanList):
        print(f"=== Scans ({scans.total}) ===")

        if not scans.items:
            print("No scans. Run the command to make one:")
            print("vulnebify run scan SCOPES -p PORTS")

        for scan in scans.items:
            scopes = scan.scopes[0]

            if len(scan.scopes) > 1:
                scopes += f" and {len(scan.scopes) - 1} more"

            started_at = scan.started_at.strftime("%Y-%m-%d %H:%M:%S")
            print(f"{scan.scan_id} - {started_at} - {scopes} - {scan.status.value}")

    def print_scan(self, scan: Scan):
        print(f"Scan ID: {scan.scan_id} ({scan.status.value})")
        print("")

        started_at = scan.started_at.strftime("%Y-%m-%d %H:%M:%S")
        if scan.ended_at:
            ended_at = scan.ended_at.strftime("%Y-%m-%d %H:%M:%S")
            duration = (scan.ended_at - scan.started_at).total_seconds()
            print(f"Duration: {started_at} - {ended_at} ({int(duration)} seconds)")
        else:
            print(f"Duration: {started_at} - running")

        scopes = ", ".join(scan.scopes)
        ports = ", ".join(scan.ports)
        scanners = ", ".join(scan.scanners)

        if scanners:
            print(
                f"Scopes: {scopes} on {ports.upper()} port(s) with {scanners} scanner(s)"
            )
        else:
            print(f"Scopes: {scopes} on {ports.upper()} port(s)")

        print("")

        if scan.hosts:
            print(f"=== Hosts ===")
            for host in scan.hosts:
                print(f"Scanned: {host.ip_str}")

        if scan.reports:
            print("")
            print(f"=== Reports ===")
            for report in scan.reports:
                print(f"Report ({report.type}): {report.slug}")

    def print_scan_initiating(
        self,
        scopes: List[str],
        ports: List[int | str],
        scanners: List[str],
    ):
        scopes = ", ".join(scopes)
        ports = ", ".join(ports)
        scanners = ", ".join(scanners)

        if scanners:
            print(
                f"🚀 Initiating scan for: {scopes} on {ports.upper()} port(s) with {scanners} scanner(s)"
            )
        else:
            print(f"🚀 Initiating scan for: {scopes} on {ports.upper()} port(s)")

    def print_scan_initiated(self, scan_id: str):
        message = f"""🆔 Scan started with ID: {scan_id}

You can check the details any time by running:
vulnebify get scan {scan_id}

🔗 Or by visiting the link: https://vulnebify.com/scan/{scan_id}
"""
        print(message)

    def print_scan_progress(self, scan: Scan, last_scanned_at: datetime) -> datetime:
        output_lines = []

        if scan.status == ScanStatus.QUEUED:
            output_lines = [f"🔄 Scan status: {scan.status.value}"]
        if scan.status == ScanStatus.RUNNING:
            ips = scan.last_scanned_ips(last_scanned_at)
            last_scanned_at = scan.last_scanned_at()

            output_lines += [f"Discovered open port(s) on {ip}" for ip in ips]

            initiated = scan.progress.initiated_tasks
            completed = scan.progress.completed_tasks
            progress_pct = (completed / initiated) * 100
            formatted_progress_pct = f"{progress_pct:.2f}"

            message = f"🔄 Scan running...Progress: {initiated}/{completed} task(s) -> {formatted_progress_pct}%"

            output_lines.append(message)
        if scan.status == ScanStatus.FINISHED:
            ips = scan.last_scanned_ips()
            duration = (scan.ended_at - scan.started_at).total_seconds()

            message = f"✅ Scan finished! Processed: {len(ips)} host(s) during {int(duration)} second(s)"

            output_lines.append(message)
        if scan.status == ScanStatus.CANCELED:
            output_lines.append(f"🛑 Scan canceled.")

        # Refresh status line (ending one)
        sys.stdout.write("\033[F\033[K")

        for line in output_lines:
            print(line)

        return last_scanned_at

    def print_scan_cancel(self, scan_id: str):
        print(f"✅ Scan {scan_id} successfully canceled!")

    def print_scanner_list(self, scanners: ScannerList):
        print(f"=== Scanners ({scanners.total}) ===")
        for scanner in scanners.items:
            print(f"{scanner.id} - {scanner.description}")

    def print_error(self, error: VulnebifyError):
        message = error.message

        if isinstance(error, VulnebifyApiError):
            if "error" in error.response and "code" in error.response["error"]:
                error_code = error.response["error"]["code"]
                message += f" Code: {error_code}."
            if "error" in error.response and "message" in error.response["error"]:
                error_message = error.response["error"]["message"]
                message += f" Message: {error_message}"

        print(message)

    def print_unexpected_error(self, error: Exception):
        print("🛑 Oh no! Unexpected error:")
        print(str(error))
        print("Create issue: https://github.com/vulnebify/vulnebify-python/issues")

    def print_exit(self):
        print("👋 Gracefully exiting. Goodbye!")


class JsonOutput(Output):
    def print_active_api_key(self, api_key: str):
        out = json.dumps(
            {
                "api_key_last4": api_key[:4] + "*" * (len(api_key) - 8) + api_key[-4:],
                "caution": "⚠️  Use Vulnebify responsibly. Only scan systems you own or have explicit permission to test. Violation may lead to legal consequences.",
                "terms_url": "https://vulnebify.com/terms",
            },
            ensure_ascii=False,
        )
        print(out)

    def print_inactive_api_key(self, api_key_hash: str):
        out = json.dumps(
            {
                "activation_url": f"https://vulnebify.com/login/?api_key_hash={api_key_hash}",
            },
        )
        print(out)

    def print_host(self, host: Host):
        print(host.model_dump_json(indent=2))

    def print_domain(self, domain: RootDomain):
        print(domain.model_dump_json(indent=2))

    def print_scan_list(self, scans: ScanList):
        print(scans.model_dump_json(indent=2))

    def print_scan(self, scan: Scan):
        print(scan.model_dump_json(indent=2))

    def print_scan_initiating(
        self,
        scopes: List[str],
        ports: List[int | str],
        scanners: List[str],
    ):
        pass

    def print_scan_initiated(self, scan_id: str):
        out = json.dumps(
            {
                "scan_id": scan_id,
                "status": ScanStatus.QUEUED.value,
                "hosts": [],
                "progress": {
                    "initiated_tasks": 0,
                    "completed_tasks": 0,
                },
            },
            indent=2,
        )
        print(out)

    def print_scan_progress(self, scan: Scan, last_scanned_at: datetime) -> datetime:
        ips = scan.last_scanned_ips(last_scanned_at)
        last_scanned_at = scan.last_scanned_at()

        out = json.dumps(
            {
                "scan_id": scan.scan_id,
                "status": scan.status,
                "hosts": ips,
                "progress": {
                    "initiated_tasks": scan.progress.initiated_tasks,
                    "completed_tasks": scan.progress.completed_tasks,
                },
            },
            indent=2,
        )
        print(out)
        return last_scanned_at

    def print_scan_cancel(self, scan_id: str):
        out = json.dumps(
            {"scan_id": scan_id, "status": ScanStatus.CANCELED.value},
            indent=2,
        )
        print(out)

    def print_scanner_list(self, scanners: ScannerList):
        print(scanners.model_dump_json(indent=2))

    def print_error(self, error: VulnebifyError):
        obj = {"error": error.message}

        if isinstance(error, VulnebifyApiError):
            obj["status_code"] = error.status_code
            obj["response"] = error.response

        out = json.dumps(obj, indent=2, ensure_ascii=False)
        print(out)

    def print_unexpected_error(self, error: Exception):
        message = f"Oh no! Unexpected error: {str(error)}"
        suggestion = (
            "Create issue: https://github.com/vulnebify/vulnebify-python/issues"
        )
        out = json.dumps({"error": message, "suggestion": suggestion}, indent=2)
        print(out)

    def print_exit(self):
        out = json.dumps(
            {"bye": "👋 Gracefully exiting. Goodbye!"}, indent=2, ensure_ascii=False
        )
        print(out)



from pydantic import BaseModel

from datetime import datetime
from typing import Dict, List, Any

from enum import Enum


class GeneratedKey(BaseModel):
    api_key: str
    api_key_hash: str

    active: bool


class ScanStatus(str, Enum):
    QUEUED = "queued"
    RUNNING = "running"
    FINALIZING = "finalizing"
    FINISHED = "finished"
    CANCELED = "canceled"


class ScanLogType(str, Enum):
    DOMAIN_SCANNED = "domain_scanned"
    HOST_SCANNED = "host_scanned"


class Label(str, Enum):
    CLOUDFLARE = "cloudflare"


class Fingerprints(BaseModel):
    sha256: str
    sha1: str
    md5: str


class Label(str, Enum):
    CLOUDFLARE = "cloudflare"


class TlsSubject(BaseModel):
    common_name: str | None = None
    organization_name: str | None = None
    country_name: str | None = None
    state_or_province_name: str | None = None
    locality_name: str | None = None


class RtspResource(BaseModel):
    url: str
    screenshot: str | None = None


class Protocol(str, Enum):
    HTTP = "http"
    RTSP = "rtsp"
    SSH = "ssh"
    FTP = "ftp"
    SMTP = "smtp"
    POP3 = "pop3"
    IMAP = "imap"
    MYSQL = "mysql"
    POSTGRESQL = "postgresql"
    REDIS = "redis"
    VNC = "vnc"
    TELNET = "telnet"
    UNKNOWN = "unknown"


class Transport(str, Enum):
    TCP = "tcp"
    UDP = "udp"


class ScanRun(BaseModel):
    scan_id: str


class ScanProgress(BaseModel):
    initiated_tasks: int = 0
    completed_tasks: int = 0


class ScanHostRef(BaseModel):
    ip_str: str

    inserted_at: datetime


class ScanReportRef(BaseModel):
    type: str
    slug: str


class Scan(BaseModel):
    scan_id: str
    status: ScanStatus

    started_at: datetime
    ended_at: datetime | None

    scopes: List[str]
    ports: List[str | int] = []
    scanners: List[str] = []

    progress: ScanProgress

    hosts: List[ScanHostRef]
    reports: List[ScanReportRef]

    def last_scanned_at(self) -> datetime:
        if not self.hosts:
            return datetime.min

        return self.hosts[-1].inserted_at

    def last_scanned_ips(self, last_scanned_at: datetime = datetime.min) -> List[str]:
        ips = []

        for host in self.hosts:
            if host.inserted_at > last_scanned_at:
                ips.append(host.ip_str)

        return ips


class ScanListItem(BaseModel):
    scan_id: str

    scopes: List[str]

    status: ScanStatus

    started_at: datetime
    ended_at: datetime | None


class ScanList(BaseModel):
    total: int
    items: List[ScanListItem]


class ScannerListItem(BaseModel):
    id: str
    description: str
    depends_on: List[str]


class ScannerList(BaseModel):
    total: int
    items: List[ScannerListItem]


class Coordinates(BaseModel):
    latitude: str
    longitude: str


class Location(BaseModel):
    country: str
    country_code: str
    city: str
    region: str

    coordinates: Coordinates


class AutonomousSystem(BaseModel):
    asn: int | None
    name: str


class Tls(BaseModel):
    subject: TlsSubject
    issuer: TlsSubject
    not_before: datetime | None
    not_after: datetime | None
    serial_number: str
    subject_alt_name: List[str | Any]
    version: int | None
    fingerprints: Fingerprints


class Http(BaseModel):
    status_code: int
    body: str
    headers: Dict[str, Any]
    fingerprints: Fingerprints


class Rtsp(BaseModel):
    resources: List[RtspResource]


class Service(BaseModel):
    port: int
    banner: str | None
    protocol: Protocol | None
    transport: Transport

    tls: Tls | None
    http: Http | None
    rtsp: Rtsp | None


class Host(BaseModel):
    ip_str: str
    ip_int: int

    first_scanned_at: datetime | None
    last_scanned_at: datetime | None

    location: Location | None
    autonomous_system: AutonomousSystem | None

    hostnames: List[str]
    cves: List[str]

    labels: List[Label]

    services: List[Service]


class DnsRecordType(str, Enum):
    A = "A"
    AAAA = "AAAA"
    CNAME = "CNAME"
    MX = "MX"
    TXT = "TXT"
    SRV = "SRV"
    NS = "NS"
    PTR = "PTR"
    SOA = "SOA"
    SPF = "SPF"
    DNSKEY = "DNSKEY"


class DnsRecord(BaseModel):
    type: DnsRecordType
    value: str


class Domain(BaseModel):
    domain: str
    dns: List[DnsRecord] = []


class RootDomain(BaseModel):
    domain: str

    first_scanned_at: datetime | None = None
    last_scanned_at: datetime | None = None

    dns: List[DnsRecord]
    subdomains: List[Domain]



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="vulnebify",
    version="1.0.1",
    packages=find_packages(exclude=["tests", "tests.*"]),
    install_requires=["requests==2.32.3", "pydantic==2.11.2"],
    entry_points={
        "console_scripts": [
            "vulnebify=vulnebify.cli:main",
        ],
    },
    author="Alex @ Vulnebify",
    author_email="contact.pypi@vulnebify.com",
    description="Python library for the Vulnebify API. Run scans, inspect hosts/domains, and fetch results with real-time updates and JSON output.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vulnebify/vulnebify-python",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



import argparse
import os
import time
import sys
import json
import getpass

from typing import List
from datetime import datetime

from vulnebify.client import Vulnebify
from vulnebify.models import Scan, ScanStatus
from vulnebify.errors import VulnebifyError
from vulnebify.output import OutputType, Output, HumanOutput, JsonOutput

CONFIG_PATH = os.path.expanduser("~/.vulnebifyrc")

VULNEBIFY_API_KEY = "VULNEBIFY_API_KEY"
VULNEBIFY_API_URL = "VULNEBIFY_API_URL"

_vulnebify: Vulnebify | None = None
_output: Output = HumanOutput()


def parse_scopes(args):
    # From file
    if args.file:
        with open(args.file) as f:
            return [line.strip() for line in f if line.strip()]

    # From positional args
    if args.scopes:
        return args.scopes

    # From stdin (pipe)
    if not sys.stdin.isatty():
        return [line.strip() for line in sys.stdin if line.strip()]

    return []


def get_api_key():
    api_key = os.getenv(VULNEBIFY_API_KEY)

    if api_key:
        return api_key

    if os.path.exists(CONFIG_PATH):
        with open(CONFIG_PATH, "r") as f:
            config = json.load(f)
            return config.get("api_key")

    return None


def save_active_api_key(api_key: str):
    with open(CONFIG_PATH, "w") as f:
        json.dump({"api_key": api_key}, f)
        _output.print_active_api_key(api_key)


def save_inactive_api_key(api_key: str, api_key_hash: str):
    with open(CONFIG_PATH, "w") as f:
        json.dump({"api_key": api_key}, f)
        _output.print_inactive_api_key(api_key_hash)


def input_api_key(api_key: str | None):
    return api_key or getpass.getpass(
        "Provide API key (key_*) to secure input. Press ENTER if you don't have one yet 🔑: "
    )


def login(api_key: str | None, api_url: str):
    api_key = input_api_key(api_key)

    if api_key:
        vulnebify = Vulnebify(api_key, api_url)

        if vulnebify.key.is_activated():
            save_active_api_key(api_key)
        else:
            message = "🛑 API key is not active. Run 'vulnebify login' again. Check --api-key argument or input value."
            raise VulnebifyError(message)

        return

    vulnebify = Vulnebify(None, api_url)  # None - without auth

    generated_key = vulnebify.key.generate()

    save_inactive_api_key(generated_key.api_key, generated_key.api_key_hash)


def run_scan(
    scopes: List[str],
    ports: List[int | str],
    scanners: List[str],
    wait: bool,
):
    if not scopes:
        message = "⚠️  You must provide at least one domain, IP, or CIDR to scan"
        raise VulnebifyError(message)

    _output.print_scan_initiating(scopes, ports, scanners)

    scan_id = _vulnebify.scan.run(scopes, ports, scanners)

    _output.print_scan_initiated(scan_id)

    if not wait:
        return

    scan: Scan = None
    last_scanned_at = datetime.min

    while True:
        scan = _vulnebify.scan.get(scan_id)

        last_scanned_at = _output.print_scan_progress(scan, last_scanned_at)

        if scan.status in [ScanStatus.FINISHED, ScanStatus.CANCELED]:
            break

        time.sleep(1)


def cancel_scan(scan_id: str):
    _vulnebify.scan.cancel(scan_id)
    _output.print_scan_cancel(scan_id)


def get_scan(scan_id: str):
    scan = _vulnebify.scan.get(scan_id)
    _output.print_scan(scan)


def list_scans():
    scans = _vulnebify.scan.list()
    _output.print_scan_list(scans)


def list_scanners():
    scanners = _vulnebify.scanner.list()
    _output.print_scanner_list(scanners)


def get_host(address: str):
    host = _vulnebify.host.get(address)
    _output.print_host(host)


def get_domain(address: str):
    domain = _vulnebify.domain.get(address)
    _output.print_domain(domain)


def print_title(parser: argparse.ArgumentParser):
    print(
        r"""
__     __ _   _  _      _   _  _____  ____   ___  _____ __   __
\ \   / /| | | || |    | \ | || ____|| __ ) |_ _||  ___|\ \ / /
 \ \ / / | | | || |    |  \| ||  _|  |  _ \  | | | |_    \ V / 
  \ V /  | |_| || |___ | |\  || |___ | |_) | | | |  _|    | |  
   \_/    \___/ |_____||_| \_||_____||____/ |___||_|      |_|            


A cyber defense platform. See more: https://about.vulnebify.com/
"""
    )
    parser.print_help()


def cli():
    # fmt: off
    
    # Shared
    output_parser = argparse.ArgumentParser(add_help=False)
    output_parser.add_argument("-o", "--output", type=OutputType, choices=["human", "json"], default=OutputType.HUMAN, help="Output format (default: human)")
    
    # Core 
    parser = argparse.ArgumentParser(prog="vulnebify")
    parser.add_argument("-a", "--api-url", default="https://api.vulnebify.com/v1", help="API url (default: https://api.vulnebify.com/v1)")
    parser.set_defaults(func=lambda _: print_title(parser))
    subparsers = parser.add_subparsers(dest="action")
    
    # LOGIN group
    login_parser = subparsers.add_parser("login", parents=[output_parser], help="Login to the API")
    login_parser.add_argument("-k", "--api-key", help="API key for authentication. Prefer using the interactive prompt for security. Only use this flag in CI/CD or trusted environments. You can also set the VULNEBIFY_API_KEY environment variable.")
    login_parser.set_defaults(func=lambda args: login(args.api_key, args.api_url))

    # RUN group
    run_parser = subparsers.add_parser("run", help="Run scan")
    run_parser.set_defaults(func=lambda _: print_title(run_parser))
    run_subparsers = run_parser.add_subparsers(dest="tool")

    # run group -> run scan
    run_scans_parser = run_subparsers.add_parser("scans", aliases=["scan"], parents=[output_parser], help="Run a scan")
    
    run_scans_parser.add_argument("scopes", nargs="*", help="Scopes to scan (domain, IP, CIDR)")
    run_scans_parser.add_argument("-f", "--file", help="Path to file with one scope per line")
    
    run_scans_parser.add_argument("-p", "--ports", nargs="*", help="Ports to scan (default: top100)")
    run_scans_parser.add_argument("-s", "--scanners", nargs="*", help="Scanners to use (default: empty)")
    run_scans_parser.add_argument("-w", "--wait", action="store_true", help="Wait for scan to finish (default: false)")
    run_scans_parser.set_defaults(func=lambda args: run_scan(parse_scopes(args), args.ports or ["top100"], args.scanners or [], args.wait))

    # CANCEL group
    cancel_parser = subparsers.add_parser("cancel", help="Cancel scan")
    cancel_parser.set_defaults(func=lambda _: print_title(cancel_parser))
    cancel_subparsers = cancel_parser.add_subparsers(dest="tool")

    # cancel group -> cancel scan
    cancel_scans_parser = cancel_subparsers.add_parser("scans", aliases=["scan"], parents=[output_parser], help="Cancel a scan")
    cancel_scans_parser.add_argument("scan_id", help="Scan ID")
    cancel_scans_parser.set_defaults(func=lambda args: cancel_scan(args.scan_id))
    
    # LIST group
    list_parser = subparsers.add_parser("list", aliases=["ls"], help="List scans, available scanners")
    list_parser.set_defaults(func=lambda _: print_title(list_parser))
    list_subparsers = list_parser.add_subparsers(help="List operations")

    # list group -> list scans
    list_scans_parser = list_subparsers.add_parser("scans", aliases=["scan"], parents=[output_parser], help="List scans")
    list_scans_parser.set_defaults(func=lambda _: list_scans())
    
    # list group -> list scanners
    list_scanners_parser = list_subparsers.add_parser("scanners", aliases=["scanner"], parents=[output_parser], help="List scanners")
    list_scanners_parser.set_defaults(func=lambda _: list_scanners())

    # GET group
    get_parser = subparsers.add_parser("get", help="Get scan, host or domain")
    get_parser.set_defaults(func=lambda _: print_title(get_parser))
    get_subparsers = get_parser.add_subparsers(help="Get operations")

    # get group -> get scan
    get_scans_parser = get_subparsers.add_parser("scans", aliases=["scan"], parents=[output_parser], help="Get scan")
    get_scans_parser.add_argument("scan_id", help="Scan ID")
    get_scans_parser.set_defaults(func=lambda args: get_scan(args.scan_id))

    # get group -> get host
    get_hosts_parser = get_subparsers.add_parser("hosts", aliases=["host"], parents=[output_parser], help="Get host")
    get_hosts_parser.add_argument("address", help="Host IPv4 / IPv6 address")
    get_hosts_parser.set_defaults(func=lambda args: get_host(args.address))

    # get group -> get domain
    get_domains_parser = get_subparsers.add_parser("domains", aliases=["domain"], parents=[output_parser], help="Get domain")
    get_domains_parser.add_argument("address", help="Domain address")
    get_domains_parser.set_defaults(func=lambda args: get_domain(args.address))

    args = parser.parse_args()
    
    if hasattr(args, "output"):
        global _output
        if args.output == OutputType.HUMAN:
            _output = HumanOutput()
        elif args.output == OutputType.JSON:
            _output = JsonOutput()
        else:
            raise NotImplementedError(f"Output `{args.output}` is not supported.")

    if args.action != "login":
        api_key = get_api_key()
        
        if not api_key:
            raise VulnebifyError("⚠️  Please log in first using 'vulnebify login'")
        
        api_url = args.api_url or os.environ.get(VULNEBIFY_API_URL)
        
        global _vulnebify
        _vulnebify = Vulnebify(api_key, api_url)
    
    args.func(args)
    # fmt: on


def main():
    try:
        cli()
    except VulnebifyError as e:
        _output.print_error(e)
    except KeyboardInterrupt:
        _output.print_exit()
    except Exception as e:
        _output.print_unexpected_error(e)


if __name__ == "__main__":
    main()
